"""A few bookkeeping functions for creating dataset files and directories.

In general we number files and directories with a suffix that is left-padded with zeros, e.g:
    dataset_0000
        - sample_000000
            sample_000000.png
            sample_000000.ply
        - sample_000001
"""

import datetime
import os


def datetime_for_filename(include_us=True, use_UTC=True) -> str:
    """Return a string with the current date and time formatted for use in filenames.

    Based on the ISO 8601 standard but optionally adds microseconds.

    Note that for the time to be accurate, make sure your system clock is synchronized.
    To check this on linux you can run: timedatectl status

    Args:
        include_us: Whether to include microseconds.
        use_UTC: Whether to use UTC time.

    Returns:
        A string with the current date and time in the format YYYY-MM-DD_HH-MM-SS-ffffff.
    """
    if use_UTC:
        now = datetime.datetime.now(datetime.timezone.utc)
    else:
        now = datetime.datetime.now()

    datetime_str = now.strftime("%Y-%m-%d_%H-%M-%S")
    if include_us:
        datetime_str += now.strftime("-%f")

    return datetime_str


def find_latest_dir(dir: str, prefix: str) -> str:
    """Find the most recent directory that starts with prefix.

    Assumes that the directories are named with the format: prefixYYYY-MM-DD_HH-MM-SS-ffffff.

    Args:
        dir: The directory where the directories are considered.
        prefix: The prefix of the directories.

    Returns:
        The most recent directory.

    """
    names = os.listdir(dir)
    sample_dirs = [name for name in names if name.startswith(prefix)]
    suffixes = [name.split(prefix)[-1] for name in sample_dirs]

    index_most_recent = suffixes.index(max(suffixes))

    sample_dir_most_recent = sample_dirs[index_most_recent]

    return os.path.join(dir, sample_dir_most_recent)


def find_highest_suffix(dir: str, name: str, extension: str | None = None) -> int:
    """Find the highest suffix of files or directories in a directory with a given name.
    Only files with exactly the same extension are considered.

    For example if the directory contains files with the names:
        sample_000000.png
        sample_000001.png
        sample_000002.png
        sample_000003.jpg
        sample_000003/
    then the highest suffix is 2.

    Args:
        dir: The directory where the files or directories are considered.
        name: The basename of the files or directories.
        extension: The extension of the files.

    Returns:
        The highest suffix found.
    """
    names = os.listdir(dir)

    # Filter by extension
    extension = "" if extension is None else extension
    names_with_same_extension = [name for name in names if os.path.splitext(name)[1] == ""]

    # Remove the extensions
    names_without_extension = [os.path.splitext(name)[0] for name in names_with_same_extension]

    # Try to split the suffixes and convert them to integers
    suffixes_int = []
    for name in names_without_extension:
        try:
            suffix = name.split("_")[-1]
            suffix = suffix[:-1].lstrip("0") + suffix[-1]  # Remove leading zeros (but keep at least one)
            suffix_int = int(suffix)
            suffixes_int.append(suffix_int)
        except ValueError:
            continue

    if len(suffixes_int) == 0:
        return -1  # No suffixes found, return -1 so that adding 1 gives 0

    return max(suffixes_int)


def ensure_dataset_dir(dataset_dir: str | None = None) -> str:
    """Creates the dataset directory if it does not exist.
    If the dataset_dir is None the name is set to "dataset_000X".
    Where X is the first available suffix

    Args:
        dataset_dir: The desired path to dataset directory.

    Returns:
        The dataset directory.
    """
    if dataset_dir is None:
        dataset_dir_name = "dataset"
        dataset_suffix_int = find_highest_suffix(".", dataset_dir_name) + 1
        dataset_dir = f"{dataset_dir_name}_{dataset_suffix_int:04d}"

    os.makedirs(dataset_dir, exist_ok=True)
    return dataset_dir



from functools import partial

import numpy as np
from airo_drake import (
    calculate_valid_joint_paths,
    concatenate_drake_trajectories,
    discretize_drake_pose_trajectory,
    time_parametrize_toppra,
)
from airo_planner import DualArmOmplPlanner, PlannerError, filter_with_distance_to_configurations, stack_joints
from airo_typing import (
    HomogeneousMatrixType,
    InverseKinematicsFunctionType,
    JointConfigurationCheckerType,
    JointConfigurationType,
    PosePathType,
    Vector3DType,
)
from loguru import logger
from pydrake.math import RigidTransform, RotationMatrix
from pydrake.multibody.plant import MultibodyPlant
from pydrake.trajectories import PiecewisePose, Trajectory


class ExhaustedOptionsError(RuntimeError):
    pass


class GraspNotFeasibleError(RuntimeError):
    pass


# TODO consider making a class for this that stores debug information
def plan_pregrasp_and_grasp_trajectory(  # noqa: C901
    planner_pregrasp: DualArmOmplPlanner,
    grasp_pose: HomogeneousMatrixType,
    start_configuration_left: JointConfigurationType,
    start_configuration_right: JointConfigurationType,
    inverse_kinematics_left_fn: InverseKinematicsFunctionType,  # same comment as for is_state_valid_fn_grasp
    inverse_kinematics_right_fn: InverseKinematicsFunctionType,
    is_state_valid_fn_grasp: JointConfigurationCheckerType,  # could make this optional and use planner's by default
    plant_toppra: MultibodyPlant,
    with_left: bool = True,
) -> tuple[Trajectory]:

    # We add 1.0 so at least one pregrasp distance fails:
    # pregrasp_distances_to_try = [0.05, 0.1, 0.15]  # , 0.2, 0.25]
    distance_min = 0.05
    distance_max = 0.25
    step = 0.01
    steps = int(np.rint((distance_max - distance_min) / step)) + 1
    pregrasp_distances_to_try = np.linspace(0.05, 0.25, steps)

    # is_state_valid_fn_grasp currently still takes a 12-DoF configuration
    def is_single_arm_state_valid_fn_grasp(joint_configuration: JointConfigurationType) -> bool:
        if with_left:
            return is_state_valid_fn_grasp(stack_joints(joint_configuration, start_configuration_right))
        else:
            return is_state_valid_fn_grasp(stack_joints(start_configuration_left, joint_configuration))

    def hardcoded_cost_fn(
        joint_configuration: JointConfigurationType,
        known_joint_configurations: list[JointConfigurationType],
        costs: list[float],
    ) -> float:
        distances = [
            np.linalg.norm(joint_configuration - known_configuration)
            for known_configuration in known_joint_configurations
        ]
        if np.min(distances) > 0.001:
            logger.warning(f"Joint configuration is not close to any known configurations. {joint_configuration} ")
        return costs[np.argmin(distances)]

    def rank_with_cost_fn(
        joint_configurations: list[JointConfigurationType], cost_fn: JointConfigurationCheckerType
    ) -> list[JointConfigurationType]:
        return sorted(joint_configurations, key=cost_fn)

    for distance in pregrasp_distances_to_try:
        logger.info(f"Planning to pregrasp pose at distance {distance}.")
        # 1. Compute pregrasp pose
        pregrasp_pose = grasp_pose.copy()
        pregrasp_pose[0:3, 3] -= distance * pregrasp_pose[0:3, 2]

        pregrasp_pose_left = pregrasp_pose if with_left else None
        pregrasp_pose_right = pregrasp_pose if not with_left else None

        # 2. Compute grasp TCP path
        rigid_transforms = [RigidTransform(pose) for pose in [pregrasp_pose, grasp_pose]]
        times = np.linspace(0, 1, len(rigid_transforms))
        pose_trajectory = PiecewisePose.MakeLinear(times=times, poses=rigid_transforms)
        grasp_tcp_path = discretize_drake_pose_trajectory(pose_trajectory).poses

        # 3 Compute valid grasp joint paths
        inverse_kinematics_fn = inverse_kinematics_left_fn if with_left else inverse_kinematics_right_fn

        grasp_path_single_arm = calculate_valid_joint_paths(
            grasp_tcp_path, inverse_kinematics_fn, is_single_arm_state_valid_fn_grasp
        )

        if len(grasp_path_single_arm) == 0:
            logger.info(f"No valid grasp joint paths found for distance {distance}, continuing to next distance.")
            continue

        if with_left:
            grasp_paths = [stack_joints(path, start_configuration_right) for path in grasp_path_single_arm]
        else:
            grasp_paths = [stack_joints(start_configuration_left, path) for path in grasp_path_single_arm]

        # Create filter function
        grasp_path_starts = [path[0] for path in grasp_paths]
        filter_fn = partial(filter_with_distance_to_configurations, joint_configurations_close=grasp_path_starts)

        # Create rank function
        grasp_trajectories = []
        times = []
        for path in grasp_paths:
            trajectory = time_parametrize_toppra(plant_toppra, path)
            times.append(trajectory.end_time())
            grasp_trajectories.append(trajectory)

        cost_fn = partial(hardcoded_cost_fn, known_joint_configurations=grasp_path_starts, costs=times)
        rank_fn = partial(rank_with_cost_fn, cost_fn=cost_fn)

        # Plan
        planner_pregrasp.filter_goal_configurations_fn = filter_fn
        planner_pregrasp.rank_goal_configurations_fn = rank_fn

        try:
            pregrasp_path = planner_pregrasp.plan_to_tcp_pose(
                start_configuration_left, start_configuration_right, pregrasp_pose_left, pregrasp_pose_right
            )
        except PlannerError as e:
            logger.info(
                f"Failed to plan to pregrasp pose at distance {distance}, continuing to next distance. Exception was:\n {e}."
            )
            continue

        pregrasp_trajectory = time_parametrize_toppra(plant_toppra, pregrasp_path)

        # # Find the grasp trajectory of which the start is closest to the pregrasp path end (=pregrasp end joints)
        # We will likely want an airo-planner helper function for this
        pregrasp_end_joints = pregrasp_path[-1]
        distances = [np.linalg.norm(start - pregrasp_end_joints) for start in grasp_path_starts]
        index_of_closest_start = np.argmin(distances)

        assert np.isclose(distances[index_of_closest_start], 0, atol=0.01)  # sanity check

        grasp_trajectory = grasp_trajectories[index_of_closest_start]

        # final set: concatenate pregrasp and grasp trajectories
        pregrasp_and_grasp_trajectory = concatenate_drake_trajectories([pregrasp_trajectory, grasp_trajectory])

        return pregrasp_and_grasp_trajectory

    raise ExhaustedOptionsError("Grasp planner exhausted all pregrasp poses to try")


def plan_to_grasp_pose_path(  # noqa: C901
    planner_pregrasp: DualArmOmplPlanner,
    grasp_pose_path: PosePathType,
    start_configuration_left: JointConfigurationType,
    start_configuration_right: JointConfigurationType,
    inverse_kinematics_left_fn: InverseKinematicsFunctionType,  # same comment as for is_state_valid_fn_grasp
    inverse_kinematics_right_fn: InverseKinematicsFunctionType,
    is_state_valid_fn_grasp: JointConfigurationCheckerType,  # could make this optional and use planner's by default
    plant_toppra: MultibodyPlant,
    with_left: bool = True,
) -> tuple[Trajectory]:

    # is_state_valid_fn_grasp currently still takes a 12-DoF configuration
    def is_single_arm_state_valid_fn_grasp(joint_configuration: JointConfigurationType) -> bool:
        if with_left:
            return is_state_valid_fn_grasp(stack_joints(joint_configuration, start_configuration_right))
        else:
            return is_state_valid_fn_grasp(stack_joints(start_configuration_left, joint_configuration))

    def hardcoded_cost_fn(
        joint_configuration: JointConfigurationType,
        known_joint_configurations: list[JointConfigurationType],
        costs: list[float],
    ) -> float:
        distances = [
            np.linalg.norm(joint_configuration - known_configuration)
            for known_configuration in known_joint_configurations
        ]
        if np.min(distances) > 0.001:
            logger.warning(f"Joint configuration is not close to any known configurations. {joint_configuration} ")
        return costs[np.argmin(distances)]

    def rank_with_cost_fn(
        joint_configurations: list[JointConfigurationType], cost_fn: JointConfigurationCheckerType
    ) -> list[JointConfigurationType]:
        return sorted(joint_configurations, key=cost_fn)

    # 3 Compute valid grasp joint paths
    inverse_kinematics_fn = inverse_kinematics_left_fn if with_left else inverse_kinematics_right_fn

    grasp_path_single_arm = calculate_valid_joint_paths(
        grasp_pose_path, inverse_kinematics_fn, is_single_arm_state_valid_fn_grasp
    )

    if len(grasp_path_single_arm) == 0:
        raise GraspNotFeasibleError("No valid joint paths found that can execute grasp.")

    if with_left:
        grasp_paths = [stack_joints(path, start_configuration_right) for path in grasp_path_single_arm]
    else:
        grasp_paths = [stack_joints(start_configuration_left, path) for path in grasp_path_single_arm]

    # Create filter function
    grasp_path_starts = [path[0] for path in grasp_paths]
    filter_fn = partial(filter_with_distance_to_configurations, joint_configurations_close=grasp_path_starts)

    # Create rank function
    grasp_trajectories = []
    times = []
    for path in grasp_paths:
        trajectory = time_parametrize_toppra(plant_toppra, path)
        times.append(trajectory.end_time())
        grasp_trajectories.append(trajectory)

    cost_fn = partial(hardcoded_cost_fn, known_joint_configurations=grasp_path_starts, costs=times)
    rank_fn = partial(rank_with_cost_fn, cost_fn=cost_fn)

    # Plan
    planner_pregrasp.filter_goal_configurations_fn = filter_fn
    planner_pregrasp.rank_goal_configurations_fn = rank_fn

    pregrasp_pose = grasp_pose_path[0]
    pregrasp_pose_left = pregrasp_pose if with_left else None
    pregrasp_pose_right = pregrasp_pose if not with_left else None

    try:
        pregrasp_path = planner_pregrasp.plan_to_tcp_pose(
            start_configuration_left, start_configuration_right, pregrasp_pose_left, pregrasp_pose_right
        )
    except PlannerError:
        raise GraspNotFeasibleError()

    pregrasp_trajectory = time_parametrize_toppra(plant_toppra, pregrasp_path)

    # # Find the grasp trajectory of which the start is closest to the pregrasp path end (=pregrasp end joints)
    # We will likely want an airo-planner helper function for this
    pregrasp_end_joints = pregrasp_path[-1]
    distances = [np.linalg.norm(start - pregrasp_end_joints) for start in grasp_path_starts]
    index_of_closest_start = np.argmin(distances)

    assert np.isclose(distances[index_of_closest_start], 0, atol=0.01)  # sanity check

    grasp_trajectory = grasp_trajectories[index_of_closest_start]

    # final set: concatenate pregrasp and grasp trajectories
    pregrasp_and_grasp_trajectory = concatenate_drake_trajectories([pregrasp_trajectory, grasp_trajectory])

    return pregrasp_and_grasp_trajectory


def make_grasp_pose_vertical(
    grasp_location: Vector3DType, gripper_open_direction: Vector3DType
) -> HomogeneousMatrixType:
    # grasp_location = np.array([0.2, 0.0, 0.7])

    gripper_forward_direction = np.array([0, 0, 1])
    Z = gripper_forward_direction / np.linalg.norm(gripper_forward_direction)
    # X = np.array([0, -1, 0])
    X = gripper_open_direction / np.linalg.norm(gripper_open_direction)
    # NOTE: we assume that the given X is perpendicular to Z

    Y = np.cross(Z, X)

    grasp_orientation = np.column_stack([X, Y, Z])
    grasp_pose_vertical = np.identity(4)
    grasp_pose_vertical[0:3, 0:3] = grasp_orientation
    grasp_pose_vertical[0:3, 3] = grasp_location
    return grasp_pose_vertical


def make_lowest_grasp_path_candidates(lowest_point: Vector3DType, grasp_depth: float = 0.0):
    grasp_pose_paths = []

    switch_angle = np.deg2rad(60)

    for angle in np.linspace(np.pi / 4, np.pi / 2, 4):
        gripper_open_direction = np.array([1, 0, 0]) if angle <= switch_angle else np.array([0, -1, 0])
        grasp_pose = make_grasp_pose_vertical(lowest_point, gripper_open_direction)

        global_y_rotation = RotationMatrix.MakeYRotation(angle).matrix()
        grasp_pose[:3, :3] = global_y_rotation @ grasp_pose[:3, :3]

        # Add grasp depth to grasp pose here
        grasp_pose[0:3, 3] += grasp_depth * grasp_pose[0:3, 2]

        for pregrasp_distance in np.linspace(0.05, 0.50, 10):
            distance = pregrasp_distance + grasp_depth
            pregrasp_pose = grasp_pose.copy()
            pregrasp_pose[0:3, 3] -= distance * grasp_pose[0:3, 2]

            rigid_transforms = [RigidTransform(pose) for pose in [pregrasp_pose, grasp_pose]]
            times = np.linspace(0, 1, len(rigid_transforms))
            pose_trajectory = PiecewisePose.MakeLinear(times=times, poses=rigid_transforms)
            steps = max(2, int(distance / 0.01))
            grasp_pose_path = discretize_drake_pose_trajectory(pose_trajectory, steps).poses

            grasp_pose_paths.append(grasp_pose_path)

    return grasp_pose_paths


def plan_lowest_point_grasp(
    lowest_point: Vector3DType,
    grasp_depth: float = 0.0,
    planner_pregrasp: DualArmOmplPlanner = None,
    home_joints_left: JointConfigurationType = None,
    home_joints_right: JointConfigurationType = None,
    inverse_kinematics_left_fn: InverseKinematicsFunctionType = None,
    inverse_kinematics_right_fn: InverseKinematicsFunctionType = None,
    is_state_valid_fn_grasp: JointConfigurationCheckerType = None,
    plant: MultibodyPlant = None,
    with_left: bool = True,
):
    grasp_pose_paths = make_lowest_grasp_path_candidates(lowest_point, grasp_depth)

    trajectory = None
    for i, grasp_pose_path in enumerate(grasp_pose_paths):
        try:
            trajectory = plan_to_grasp_pose_path(
                planner_pregrasp,
                grasp_pose_path,
                home_joints_left,
                home_joints_right,
                inverse_kinematics_left_fn,
                inverse_kinematics_right_fn,
                is_state_valid_fn_grasp,
                plant,
                with_left=with_left,
            )
            logger.success(f"Grasp pose path {i} is feasible.")
            break
        except GraspNotFeasibleError:
            logger.info(f"Grasp pose path {i} is not feasible.")
            continue

    if trajectory is None:
        raise ExhaustedOptionsError("All lowest point grasp paths are infeasible.")
    return trajectory



import numpy as np

TCP_TRANSFORM = np.identity(4)
TCP_TRANSFORM[2, 3] = 0.175

JOINT_BOUNDS_LOWER = np.deg2rad([-360, -195, -160, -360, -360, -360])
JOINT_BOUNDS_UPPER = np.deg2rad([360, 15, 160, 360, 360, 360])
JOINT_BOUNDS = JOINT_BOUNDS_LOWER, JOINT_BOUNDS_UPPER



from typing import Tuple

from airo_typing import BoundingBox3DType

BBOX_CLOTH_ON_TABLE = (-0.55, -0.25, 0.01), (0.55, 0.25, 0.2)
BBOX_CLOTH_IN_THE_AIR = (-0.2, -0.2, 0.2), (0.2, 0.2, 0.8)


def bbox_to_mins_and_sizes(
    bounding_box: BoundingBox3DType,
) -> Tuple[Tuple[float, float, float], Tuple[float, float, float]]:
    """Convert a bounding box specified as the mix and max corners to the alternative mins and sizes format.
    This is the format used by Rerun.

    Args:
        bounding_box: bounding box specified as the mix and max corners

    Returns:
        bounding box specified as a tuple: the mins and sizes
    """
    mins, maxs = bounding_box
    sizes = (maxs[0] - mins[0], maxs[1] - mins[1], maxs[2] - mins[2])
    return (mins, sizes)



from typing import Tuple

import cv2
import numpy as np
from airo_camera_toolkit.pinhole_operations.projection import project_points_to_image_plane
from airo_spatial_algebra import SE3Container, transform_points
from airo_typing import CameraIntrinsicsMatrixType, HomogeneousMatrixType, OpenCVIntImageType, Vector3DType


def draw_point_3d(
    image: OpenCVIntImageType,
    point_3d: Vector3DType,
    intrinsics: CameraIntrinsicsMatrixType,
    camera_pose: HomogeneousMatrixType,
    color: Tuple[int, int, int],
) -> None:
    """Draws a 3D point on an image. Image and color may be in either RGB or BGR.

    Args:
        image: The image to draw on.
        point_3d: The 3D point to draw.
        color: The color of the point.
    """
    X_C_W = np.linalg.inv(camera_pose)
    point_2d = project_points_to_image_plane(transform_points(X_C_W, point_3d), intrinsics).squeeze()
    point_2d_int = np.rint(point_2d).astype(int)
    cv2.circle(image, tuple(point_2d_int), 10, color, thickness=2)


def draw_pose(
    image: OpenCVIntImageType,
    pose: HomogeneousMatrixType,
    intrinsics: CameraIntrinsicsMatrixType,
    camera_pose: HomogeneousMatrixType,
    length: float = 0.1,
) -> None:
    """Draws a pose on an image. Image must be BGR to have correct axes colors.

    Args:
        image: The image to draw on.
        pose: The pose to draw.
        color: The color of the pose.
    """
    X_W_C = camera_pose
    X_W_P = pose
    X_C_P = np.linalg.inv(X_W_C) @ X_W_P
    frame_pose_in_camera = X_C_P

    # copied from airo_camera_toolkit/calibration/fiducial_markers.py
    charuco_se3 = SE3Container.from_homogeneous_matrix(frame_pose_in_camera)
    rvec = charuco_se3.orientation_as_rotation_vector
    tvec = charuco_se3.translation
    cv2.drawFrameAxes(image, intrinsics, None, rvec, tvec, length)



from airo_camera_toolkit.point_clouds.operations import crop_point_cloud
from cloth_tools.point_clouds.camera import get_image_and_filtered_point_cloud


def sufficient_points_in_bbox(station, bbox, minimum_amount_of_points=1000):
    camera = station.camera
    camera_pose = station.camera_pose

    _, _, point_cloud = get_image_and_filtered_point_cloud(camera, camera_pose)
    point_cloud_cropped = crop_point_cloud(point_cloud, bbox)

    return len(point_cloud_cropped.points) > minimum_amount_of_points



from typing import Tuple

from airo_camera_toolkit.cameras.zed.zed2i import Zed2i
from airo_camera_toolkit.point_clouds.conversions import open3d_to_point_cloud, point_cloud_to_open3d
from airo_camera_toolkit.point_clouds.operations import filter_point_cloud
from airo_typing import HomogeneousMatrixType, NumpyDepthMapType, NumpyIntImageType, PointCloud


def get_image_and_filtered_point_cloud(
    camera: Zed2i, camera_pose: HomogeneousMatrixType
) -> Tuple[NumpyIntImageType, NumpyDepthMapType, PointCloud]:
    """Get an RGB image and a point cloud from the camera.
    Low confidence points are filtered out and the point cloud is transformed to the world frame.

    Args:
        camera: The camera.
        camera_pose: The pose of the camera in the world frame.

    Returns:
        image_rgb: The RGB image from the camera.
        point_cloud_filtered: The point cloud in the world frame.
    """
    image_rgb = camera.get_rgb_image_as_int()
    point_cloud_in_camera = camera._retrieve_colored_point_cloud()
    depth_map = camera._retrieve_depth_map()
    confidence_map = camera._retrieve_confidence_map()

    # Transform the point cloud to the world frame
    pcd_in_camera = point_cloud_to_open3d(point_cloud_in_camera)
    pcd = pcd_in_camera.transform(camera_pose)  # transform to world frame (= base frame of left robot)

    # Filter the low confidence points
    confidence_mask = (confidence_map <= 1.0).reshape(-1)  # Threshold and flatten

    point_cloud = open3d_to_point_cloud(pcd)
    point_cloud_filtered = filter_point_cloud(point_cloud, confidence_mask.nonzero())
    return image_rgb, depth_map, point_cloud_filtered



import numpy as np
from airo_robots.manipulators.bimanual_position_manipulator import DualArmPositionManipulator
from pydrake.trajectories import Trajectory


def ensure_dual_arm_at_joint_configuration(dual_arm, joints_left, joints_right, tolerance=0.1) -> None:
    """Sanity check that the arm are were you expect them to be,
    e.g. close to that start of a path you are about to execute.

    Raises ValueError if the arms are not at the expected joints.
    """
    current_joints_left = dual_arm.left_manipulator.get_joint_configuration()
    current_joints_right = dual_arm.right_manipulator.get_joint_configuration()

    left_distance = np.linalg.norm(current_joints_left - joints_left)
    right_distance = np.linalg.norm(current_joints_right - joints_right)

    if left_distance > tolerance:
        raise ValueError(
            f"Left arm is at {current_joints_left} but should be at {joints_left}, distance: {left_distance}"
        )
    if right_distance > tolerance:
        raise ValueError(
            f"Right arm is at {current_joints_right} but should be at {joints_right}, distance: {right_distance}"
        )


def execute_dual_arm_drake_trajectory(dual_arm: DualArmPositionManipulator, joint_trajectory: Trajectory):
    # TODO don't receive joint and time trajectory separately, but as a single PathParametrizedTrajectory
    # TODO create and use discretize_dual_arm_trajectory
    assert joint_trajectory.start_time() == 0.0

    start_joints = joint_trajectory.value(0).squeeze()
    start_joints_left = start_joints[0:6]
    start_joints_right = start_joints[6:12]

    ensure_dual_arm_at_joint_configuration(dual_arm, start_joints_left, start_joints_right)

    period = 0.005
    duration = joint_trajectory.end_time()

    n_servos = int(np.ceil(duration / period))
    period_adjusted = duration / n_servos  # can be slightly different from period due to rounding

    for t in np.linspace(0, duration, n_servos):
        joints = joint_trajectory.value(t).squeeze()
        joints_left = joints[0:6]
        joints_right = joints[6:12]
        left_servo = dual_arm.left_manipulator.servo_to_joint_configuration(joints_left, period_adjusted)
        right_servo = dual_arm.right_manipulator.servo_to_joint_configuration(joints_right, period_adjusted)
        left_servo.wait()
        right_servo.wait()

    # This avoids the abrupt stop and "thunk" sounds at the end of paths that end with non-zero velocity
    # However, I believe these functions are blocking, so right only stops after left has stopped.
    dual_arm.left_manipulator.rtde_control.servoStop(2.0)
    dual_arm.right_manipulator.rtde_control.servoStop(2.0)

    left_finished = dual_arm.left_manipulator.move_to_joint_configuration(joints_left)
    right_finished = dual_arm.right_manipulator.move_to_joint_configuration(joints_right)

    left_finished.wait()
    right_finished.wait()



from collections import deque

import scipy
from airo_typing import WrenchType


class WrenchSmoother:
    def __init__(self, history: int):
        self.history = history
        self.sigma = 2 * history + 1

        self.forces_x = deque(maxlen=history)
        self.forces_y = deque(maxlen=history)
        self.forces_z = deque(maxlen=history)
        self.torques_x = deque(maxlen=history)
        self.torques_y = deque(maxlen=history)
        self.torques_z = deque(maxlen=history)

        self.wrench_deques = [
            self.forces_x,
            self.forces_y,
            self.forces_z,
            self.torques_x,
            self.torques_y,
            self.torques_z,
        ]

    def smooth(self, wrench_deque: deque) -> float:
        return scipy.ndimage.gaussian_filter1d(wrench_deque, sigma=self.sigma)[-1]

    def smooth_wrench(self) -> WrenchType:
        wrench = []

        for wrench_deque in self.wrench_deques:
            wrench.append(self.smooth(wrench_deque))

        return wrench

    def add_wrench(self, wrench: WrenchType) -> WrenchType:
        for i, component in enumerate(wrench):
            self.wrench_deques[i].append(component)

        return self.smooth_wrench()



import os
import time
from pathlib import Path
from typing import Tuple

from airo_dataset_tools.data_parsers.pose import Pose
from airo_robots.grippers.hardware.robotiq_2f85_urcap import Robotiq2F85
from airo_robots.manipulators.bimanual_position_manipulator import DualArmPositionManipulator
from airo_robots.manipulators.hardware.ur_rtde import URrtde
from airo_typing import HomogeneousMatrixType

CONFIG_DIR = "CLOTH_TOOLS_CONFIG_DIR"


def get_config_dir() -> str:
    """Path to a directory with configuration files such as camera extrinsics.

    Set in your conda with like so:
    conda env config vars set CLOTH_TOOLS_CONFIG_DIR=/home/victor/cloth-competition/config/

    Returns:
        str: Path to the config directory.
    """
    if CONFIG_DIR not in os.environ:
        raise ValueError(f"Environment variable {CONFIG_DIR} not set")
    return os.environ[CONFIG_DIR]


def load_camera_pose_in_left_and_right() -> Tuple[HomogeneousMatrixType, HomogeneousMatrixType]:
    """Load the camera pose in the left and right robot's base frame.

    Returns:
        camera pose in left and right robot's base frame
    """
    with open(Path(get_config_dir()) / "camera_pose_in_left.json", "r") as f:
        camera_pose_in_left = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(Path(get_config_dir()) / "camera_pose_in_right.json", "r") as f:
        camera_pose_in_right = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    return camera_pose_in_left, camera_pose_in_right


def setup_dual_arm_ur5e_in_world(
    X_W_LCB: HomogeneousMatrixType,
    X_W_RCB: HomogeneousMatrixType,
    ip_address_left: str = "10.42.0.163",
    ip_address_right: str = "10.42.0.162",
) -> DualArmPositionManipulator:
    """Connect to the UR5e robots and Robotiq 2F-85 grippers. Sets the world frame to the base frame of the left robot.

    Returns:
        The initialized dual arm.
    """
    max_attempts = 3
    for i in range(max_attempts):
        try:
            gripper_left = Robotiq2F85(ip_address_left)
            robot_left = URrtde(ip_address_left, URrtde.UR3E_CONFIG, gripper_left)

            gripper_right = Robotiq2F85(ip_address_right)
            robot_right = URrtde(ip_address_right, URrtde.UR3E_CONFIG, gripper_right)

            dual_arm = DualArmPositionManipulator(robot_left, X_W_LCB, robot_right, X_W_RCB)
            return dual_arm
        except RuntimeError as e:
            print(f"Failed to connect to UR5e robots. Attempt {i+1}/{max_attempts}. Error:\n {e}")
            if i == max_attempts - 1:
                raise e
            time.sleep(2)



from airo_camera_toolkit.cameras.zed.zed2i import Zed2i
from airo_robots.manipulators.bimanual_position_manipulator import DualArmPositionManipulator
from airo_typing import CameraExtrinsicMatrixType


class DualArmStation:
    """A station is a collection of hardware, e.g. the robots, camera, but also configuration such camera extrinsics,
    or can even contain a simulation environment for motion planning etc.

    In this project we assume a dual arm station contains at least:
    * two robot arms (grouped into a DualArmPositionManipulator object)
    * a stereo RGBD camera (for now must be a ZED2i camera)
    * the pose of the camera in the world frame.

    The world frame is defined to be the base frame of the left robot.
    """

    def __init__(self, dual_arm: DualArmPositionManipulator, camera: Zed2i, camera_pose: CameraExtrinsicMatrixType):
        self.dual_arm = dual_arm
        self.camera = camera
        self.camera_pose = camera_pose



import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import List

import cv2
import numpy as np
import open3d as o3d
from airo_camera_toolkit.point_clouds.conversions import open3d_to_point_cloud, point_cloud_to_open3d
from airo_camera_toolkit.utils.image_converter import ImageConverter
from airo_dataset_tools.data_parsers.camera_intrinsics import CameraIntrinsics
from airo_dataset_tools.data_parsers.pose import Pose
from airo_typing import (
    CameraExtrinsicMatrixType,
    CameraIntrinsicsMatrixType,
    CameraResolutionType,
    HomogeneousMatrixType,
    JointConfigurationType,
    NumpyDepthMapType,
    NumpyIntImageType,
    PointCloud,
)
from loguru import logger
from pydantic import BaseModel


class JointConfigurationModel(BaseModel):
    values: List[float]


@dataclass
class CompetitionObservation:
    # Data
    image_left: NumpyIntImageType
    image_right: NumpyIntImageType
    depth_map: NumpyDepthMapType  # Depth should be aligned with the left image
    point_cloud: PointCloud  # Expressed in the world frame
    depth_image: NumpyIntImageType  # Depth image for visualization
    confidence_map: NumpyDepthMapType  # Confidence of the depth map as returned by the ZED SDK

    # Poses
    camera_pose_in_world: CameraExtrinsicMatrixType
    arm_left_pose_in_world: HomogeneousMatrixType
    arm_right_pose_in_world: HomogeneousMatrixType
    arm_left_joints: JointConfigurationType
    arm_right_joints: JointConfigurationType
    arm_left_tcp_pose_in_world: HomogeneousMatrixType
    arm_right_tcp_pose_in_world: HomogeneousMatrixType
    right_camera_pose_in_left_camera: HomogeneousMatrixType

    # Camera intrinsics
    camera_intrinsics: CameraIntrinsicsMatrixType
    camera_resolution: CameraResolutionType


# We put these in a dictionary so both the saving and loading code can use the same file names and extensions.
COMPETITION_OBSERVATION_FILENAMES = {
    "image_left": "image_left.png",
    "image_right": "image_right.png",
    "depth_map": "depth_map.tiff",
    "point_cloud": "point_cloud.ply",
    "depth_image": "depth_image.jpg",
    "confidence_map": "confidence_map.tiff",
    "camera_pose_in_world": "camera_pose_in_world.json",
    "arm_left_pose_in_world": "arm_left_pose_in_world.json",
    "arm_right_pose_in_world": "arm_right_pose_in_world.json",
    "arm_left_joints": "arm_left_joints.json",
    "arm_right_joints": "arm_right_joints.json",
    "arm_left_tcp_pose_in_world": "arm_left_tcp_pose_in_world.json",
    "arm_right_tcp_pose_in_world": "arm_right_tcp_pose_in_world.json",
    "right_camera_pose_in_left_camera": "right_camera_pose_in_left_camera.json",
    "camera_intrinsics": "camera_intrinsics.json",  # Resolution is saved together with the intrinsics
}


def save_competition_observation(observation: CompetitionObservation, observation_dir: str) -> None:
    os.makedirs(observation_dir, exist_ok=True)

    # Turn filenames into filepaths
    filepaths = {
        key: str(Path(observation_dir) / filename) for key, filename in COMPETITION_OBSERVATION_FILENAMES.items()
    }

    # Convert images from RGB to BGR
    image_left = ImageConverter.from_numpy_int_format(observation.image_left).image_in_opencv_format
    image_right = ImageConverter.from_numpy_int_format(observation.image_right).image_in_opencv_format
    depth_image = ImageConverter.from_numpy_int_format(observation.depth_image).image_in_opencv_format

    cv2.imwrite(filepaths["image_left"], image_left)
    cv2.imwrite(filepaths["image_right"], image_right)
    cv2.imwrite(filepaths["depth_image"], depth_image)
    cv2.imwrite(filepaths["depth_map"], observation.depth_map)
    cv2.imwrite(filepaths["confidence_map"], observation.confidence_map)

    pcd = point_cloud_to_open3d(observation.point_cloud)

    if pcd.point.colors.numpy().dtype != np.uint8:
        logger.warning("Point cloud colors were not uint8, converting to uint8.")
        pcd.point.colors = o3d.utility.Vector3dVector((pcd.point.colors.numpy() * 255.0).astype(np.uint8))

    o3d.t.io.write_point_cloud(filepaths["point_cloud"], pcd)

    with open(filepaths["camera_intrinsics"], "w") as f:
        intrinsics_model_left = CameraIntrinsics.from_matrix_and_resolution(
            observation.camera_intrinsics, observation.camera_resolution
        )
        json.dump(intrinsics_model_left.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["camera_pose_in_world"], "w") as f:
        camera_pose_model = Pose.from_homogeneous_matrix(observation.camera_pose_in_world)
        json.dump(camera_pose_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["arm_left_pose_in_world"], "w") as f:
        arm_left_pose_model = Pose.from_homogeneous_matrix(observation.arm_left_pose_in_world)
        json.dump(arm_left_pose_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["arm_right_pose_in_world"], "w") as f:
        arm_right_pose_model = Pose.from_homogeneous_matrix(observation.arm_right_pose_in_world)
        json.dump(arm_right_pose_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["arm_left_joints"], "w") as f:
        joints_model = JointConfigurationModel(values=list(observation.arm_left_joints))
        json.dump(joints_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["arm_right_joints"], "w") as f:
        joints_model = JointConfigurationModel(values=list(observation.arm_right_joints))
        json.dump(joints_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["arm_left_tcp_pose_in_world"], "w") as f:
        arm_left_tcp_pose_model = Pose.from_homogeneous_matrix(observation.arm_left_tcp_pose_in_world)
        json.dump(arm_left_tcp_pose_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["arm_right_tcp_pose_in_world"], "w") as f:
        arm_right_tcp_pose_model = Pose.from_homogeneous_matrix(observation.arm_right_tcp_pose_in_world)
        json.dump(arm_right_tcp_pose_model.model_dump(exclude_none=True), f, indent=4)

    with open(filepaths["right_camera_pose_in_left_camera"], "w") as f:
        right_camera_pose_model = Pose.from_homogeneous_matrix(observation.right_camera_pose_in_left_camera)
        json.dump(right_camera_pose_model.model_dump(exclude_none=True), f, indent=4)

    logger.info(f"Saved observation to {observation_dir}")


def load_competition_observation(observation_dir: str) -> CompetitionObservation:
    """Loads a competition observation from a directory.

    Args:
        observation_dir: The directory containing the observation files.

    Returns:
        A CompetitionObservation instance.
    """
    # Turn filenames into filepaths
    filepaths = {
        key: str(Path(observation_dir) / filename) for key, filename in COMPETITION_OBSERVATION_FILENAMES.items()
    }

    image_left = cv2.imread(filepaths["image_left"])
    image_right = cv2.imread(filepaths["image_right"])
    depth_image = cv2.imread(filepaths["depth_image"])
    depth_map = cv2.imread(filepaths["depth_map"], cv2.IMREAD_ANYDEPTH)
    confidence_map = cv2.imread(filepaths["confidence_map"], cv2.IMREAD_ANYDEPTH)

    # Convert images from BGR to RGB
    image_left = ImageConverter.from_opencv_format(image_left).image_in_numpy_int_format
    image_right = ImageConverter.from_opencv_format(image_right).image_in_numpy_int_format
    depth_image = ImageConverter.from_opencv_format(
        depth_image
    ).image_in_numpy_int_format  # in case it's not grayscale

    with open(filepaths["camera_pose_in_world"], "r") as f:
        camera_pose_in_world = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(filepaths["arm_left_pose_in_world"], "r") as f:
        arm_left_pose_in_world = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(filepaths["arm_right_pose_in_world"], "r") as f:
        arm_right_pose_in_world = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(filepaths["arm_left_joints"], "r") as f:
        arm_left_joints = np.array(JointConfigurationModel.model_validate_json(f.read()).values)

    with open(filepaths["arm_right_joints"], "r") as f:
        arm_right_joints = np.array(JointConfigurationModel.model_validate_json(f.read()).values)

    with open(filepaths["arm_left_tcp_pose_in_world"], "r") as f:
        arm_left_tcp_pose_in_world = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(filepaths["arm_right_tcp_pose_in_world"], "r") as f:
        arm_right_tcp_pose_in_world = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(filepaths["right_camera_pose_in_left_camera"], "r") as f:
        right_camera_pose_in_left_camera = Pose.model_validate_json(f.read()).as_homogeneous_matrix()

    with open(filepaths["camera_intrinsics"], "r") as f:
        intrinsics_model = CameraIntrinsics.model_validate_json(f.read())
        camera_intrinsics = intrinsics_model.as_matrix()
        camera_resolution = intrinsics_model.image_resolution.as_tuple()

    pcd = o3d.t.io.read_point_cloud(filepaths["point_cloud"])
    point_cloud = open3d_to_point_cloud(pcd)

    return CompetitionObservation(
        image_left=image_left,
        image_right=image_right,
        depth_map=depth_map,
        point_cloud=point_cloud,
        depth_image=depth_image,
        confidence_map=confidence_map,
        camera_pose_in_world=camera_pose_in_world,
        arm_left_pose_in_world=arm_left_pose_in_world,
        arm_right_pose_in_world=arm_right_pose_in_world,
        arm_left_joints=arm_left_joints,
        arm_right_joints=arm_right_joints,
        arm_left_tcp_pose_in_world=arm_left_tcp_pose_in_world,
        arm_right_tcp_pose_in_world=arm_right_tcp_pose_in_world,
        right_camera_pose_in_left_camera=right_camera_pose_in_left_camera,
        camera_intrinsics=camera_intrinsics,
        camera_resolution=camera_resolution,
    )



import rerun as rr
from airo_camera_toolkit.interfaces import RGBCamera
from airo_typing import HomogeneousMatrixType


def rr_log_camera(camera: RGBCamera, camera_pose: HomogeneousMatrixType, entity_path: str = "world/camera") -> None:
    """Log the camera and its pose to Rerun.

    Args:
        camera: The camera.
        camera_pose: The pose of the camera in the world frame.
    """
    rr_pinhole = rr.Pinhole(image_from_camera=camera.intrinsics_matrix(), resolution=camera.resolution)
    rr_camera_pose = rr.Transform3D(translation=camera_pose[0:3, 3], mat3x3=camera_pose[0:3, 0:3])
    rr.log(entity_path, rr_pinhole)
    rr.log(entity_path, rr_camera_pose)



from airo_camera_toolkit.interfaces import StereoRGBDCamera
from airo_camera_toolkit.point_clouds.conversions import open3d_to_point_cloud, point_cloud_to_open3d
from cloth_tools.dataset.format import CompetitionObservation
from cloth_tools.stations.competition_station import CompetitionStation


def collect_observation(station: CompetitionStation) -> CompetitionObservation:
    camera = station.camera

    camera_pose_in_world = station.camera_pose
    arm_left_pose_in_world = station.left_arm_pose
    arm_right_pose_in_world = station.right_arm_pose
    right_camera_pose_in_left_camera = camera.pose_of_right_view_in_left_view

    camera_intrinsics = camera.intrinsics_matrix()
    camera_resolution = camera.resolution

    image_left = camera.get_rgb_image_as_int()
    image_right = camera._retrieve_rgb_image_as_int(StereoRGBDCamera.RIGHT_RGB)
    depth_map = camera._retrieve_depth_map()
    depth_image = camera._retrieve_depth_image()
    confidence_map = camera._retrieve_confidence_map()

    point_cloud_in_camera = camera._retrieve_colored_point_cloud()
    pcd_in_camera = point_cloud_to_open3d(point_cloud_in_camera)
    pcd = pcd_in_camera.transform(camera_pose_in_world)  # transform to world frame (= base frame of left robot)
    point_cloud = open3d_to_point_cloud(pcd)

    arm_left_tcp_pose_in_base = station.dual_arm.left_manipulator.get_tcp_pose()
    arm_left_joints = station.dual_arm.left_manipulator.get_joint_configuration()
    arm_right_tcp_pose_in_base = station.dual_arm.right_manipulator.get_tcp_pose()
    arm_right_joints = station.dual_arm.right_manipulator.get_joint_configuration()

    # Convert TCP poses from robot base frames to world frame
    X_LCB_TCPL = arm_left_tcp_pose_in_base
    X_RCB_TCPR = arm_right_tcp_pose_in_base

    X_W_LCB = arm_left_pose_in_world
    X_W_RCB = arm_right_pose_in_world

    X_W_TCPL = X_W_LCB @ X_LCB_TCPL
    X_W_TCPR = X_W_RCB @ X_RCB_TCPR

    arm_left_tcp_pose_in_world = X_W_TCPL
    arm_right_tcp_pose_in_world = X_W_TCPR

    observation = CompetitionObservation(
        image_left=image_left,
        image_right=image_right,
        depth_map=depth_map,
        point_cloud=point_cloud,
        depth_image=depth_image,
        confidence_map=confidence_map,
        camera_pose_in_world=camera_pose_in_world,
        arm_left_pose_in_world=arm_left_pose_in_world,
        arm_right_pose_in_world=arm_right_pose_in_world,
        arm_left_joints=arm_left_joints,
        arm_right_joints=arm_right_joints,
        arm_left_tcp_pose_in_world=arm_left_tcp_pose_in_world,
        arm_right_tcp_pose_in_world=arm_right_tcp_pose_in_world,
        right_camera_pose_in_left_camera=right_camera_pose_in_left_camera,
        camera_intrinsics=camera_intrinsics,
        camera_resolution=camera_resolution,
    )
    return observation



import multiprocessing
from functools import partial

import numpy as np
from airo_camera_toolkit.cameras.multiprocess.multiprocess_stereo_rgbd_camera import (
    MultiprocessStereoRGBDPublisher,
    MultiprocessStereoRGBDReceiver,
)
from airo_camera_toolkit.cameras.zed.zed2i import Zed2i
from airo_camera_toolkit.image_transforms.transforms.crop import Crop
from airo_camera_toolkit.interfaces import StereoRGBDCamera
from airo_drake import DualArmScene, add_meshcat, finish_build
from airo_planner import DualArmOmplPlanner, DualArmPlanner
from cloth_tools.config import load_camera_pose_in_left_and_right, setup_dual_arm_ur5e_in_world
from cloth_tools.drake.scenes import add_cloth_competition_dual_ur5e_scene
from cloth_tools.kinematics.constants import JOINT_BOUNDS, TCP_TRANSFORM
from cloth_tools.kinematics.inverse_kinematics import inverse_kinematics_in_world_fn
from cloth_tools.stations.coordinate_frames import create_egocentric_world_frame
from cloth_tools.stations.dual_arm_station import DualArmStation
from loguru import logger
from pydrake.planning import RobotDiagramBuilder, SceneGraphCollisionChecker


def check_zed_point_cloud_completeness(camera: Zed2i):
    # Check whether the point cloud is complete, i.e. if there are any points closers than 1.0 meters
    point_cloud = camera.get_colored_point_cloud()
    image_rgb = camera.get_rgb_image_as_int()
    image_right_rgb = camera._retrieve_rgb_image_as_int(StereoRGBDCamera.RIGHT_RGB)
    confidence_map = camera._retrieve_confidence_map()
    depth_map = camera._retrieve_depth_map()
    depth_image = camera._retrieve_depth_image()

    distances = np.linalg.norm(point_cloud.points, axis=1)
    if not np.any(distances < 1.2):
        print(distances)
        logger.info("The point cloud is not complete, logging it to rerun.")
        import rerun as rr

        rr.init("Competition Station - Point cloud", spawn=True)
        rr.log("world/point_cloud", rr.Points3D(positions=point_cloud.points, colors=point_cloud.colors))
        rr.log("image", rr.Image(image_rgb).compress(jpeg_quality=90))
        rr.log("image_right", rr.Image(image_right_rgb).compress(jpeg_quality=90))
        rr.log("depth_image", rr.Image(depth_image).compress(jpeg_quality=90))
        rr.log("depth_map", rr.DepthImage(depth_map))
        rr.log("confidence_map", rr.Image(confidence_map))

        raise RuntimeError("The point cloud is incomplete. Restart the ZED2i camera.")


class CompetitionStation(DualArmStation):
    """
    This station specifically contains the hardware setup for the ICRA 2024 competition.
    It (currently) consists of two UR5e robots and a single ZED2i camera.
    The robots are mounted approximately 0.9 meter apart.
    """

    def __init__(self, create_multiprocess_camera: bool = True) -> None:

        # Setting up the camera
        camera_kwargs = {
            "resolution": Zed2i.RESOLUTION_2K,
            "depth_mode": Zed2i.NEURAL_DEPTH_MODE,
            "fps": 15,
        }
        self.camera_publisher = None

        for i in range(5):
            logger.info(f"Initializing camera... (attempt: {i})")
            if create_multiprocess_camera:
                if i == 0:
                    multiprocessing.set_start_method("spawn")

                # Running the camera in a seperate process enables us to record videos even if the main process is blocking
                self.camera_publisher = MultiprocessStereoRGBDPublisher(Zed2i, camera_kwargs)
                self.camera_publisher.start()
                camera = MultiprocessStereoRGBDReceiver("camera")
            else:
                camera = Zed2i(**camera_kwargs)

            try:
                check_zed_point_cloud_completeness(camera)
                break
            except RuntimeError as e:
                logger.error("Failed to initialize camera. Retrying...")
                logger.warning(e)
                if self.camera_publisher is not None:
                    self.camera_publisher.stop()
                    self.camera_publisher.join()
                    del camera
                if i == 4:
                    raise e

        # Image crop used to check motion blur of hanging cloth
        # Note that parts of the cloth may be outside the crop
        image_rgb = camera.get_rgb_image_as_int()
        self.hanging_cloth_crop = Crop(image_rgb.shape, x=1000, y=200, w=200, h=500)

        camera_pose_in_left, camera_pose_in_right = load_camera_pose_in_left_and_right()

        X_W_C, X_W_LCB, X_W_RCB = create_egocentric_world_frame(camera_pose_in_left, camera_pose_in_right)

        camera_pose = X_W_C  # this must be consistent with the setup_dual_arm_ur5e call below
        self.left_arm_pose = X_W_LCB
        self.right_arm_pose = X_W_RCB

        dual_arm = setup_dual_arm_ur5e_in_world(X_W_LCB, X_W_RCB)
        super().__init__(dual_arm, camera, camera_pose)

        # Setting some home joints
        self.home_joints_left = np.deg2rad([180, -120, 60, -30, -90, -90])
        self.home_joints_right = np.deg2rad([-180, -60, -60, -150, 90, 90])

        self.joint_bounds_left = JOINT_BOUNDS
        self.joint_bounds_right = JOINT_BOUNDS

        # Planner for the two arms without obstacles (only the table)
        self.planner: DualArmPlanner = self._setup_planner(X_W_LCB, X_W_RCB)

        # Publishing the current joint is purely for debugging. This way you can check in meshcat if the robot is
        # mounted the same way as in the real world as in the simulation.
        robot_diagram = self.drake_scene.robot_diagram
        context = robot_diagram.CreateDefaultContext()
        current_joints_left = dual_arm.left_manipulator.get_joint_configuration()
        current_joints_right = dual_arm.right_manipulator.get_joint_configuration()
        plant = robot_diagram.plant()
        plant_context = plant.GetMyContextFromRoot(context)
        plant.SetPositions(plant_context, self.drake_scene.arm_left_index, current_joints_left)
        plant.SetPositions(plant_context, self.drake_scene.arm_right_index, current_joints_right)
        robot_diagram.ForcedPublish(context)

        logger.success("CompetitionStation initialized.")

    def __del__(self):
        if self.camera_publisher is not None:
            self.camera_publisher.stop()
            self.camera_publisher.join()

    def _setup_planner(self, X_W_LCB, X_W_RCB) -> DualArmOmplPlanner:
        # Creating the default scene

        robot_diagram_builder = RobotDiagramBuilder()

        meshcat = add_meshcat(robot_diagram_builder)
        meshcat.SetCameraPose([-1.5, 0, 1.0], [0, 0, 0])

        (arm_left_index, arm_right_index), (
            gripper_left_index,
            gripper_right_index,
        ) = add_cloth_competition_dual_ur5e_scene(robot_diagram_builder, X_W_LCB, X_W_RCB)
        robot_diagram, _ = finish_build(robot_diagram_builder, meshcat)

        scene = DualArmScene(
            robot_diagram, arm_left_index, arm_right_index, gripper_left_index, gripper_right_index, meshcat
        )
        self.drake_scene = scene

        collision_checker = SceneGraphCollisionChecker(
            model=scene.robot_diagram,
            robot_model_instances=[
                scene.arm_left_index,
                scene.arm_right_index,
                scene.gripper_left_index,
                scene.gripper_right_index,
            ],
            edge_step_size=0.125,  # Arbitrary value: we don't use the CheckEdgeCollisionFree
            env_collision_padding=0.005,
            self_collision_padding=0.005,
        )

        inverse_kinematics_left_fn = partial(
            inverse_kinematics_in_world_fn, X_W_CB=X_W_LCB, tcp_transform=TCP_TRANSFORM
        )
        inverse_kinematics_right_fn = partial(
            inverse_kinematics_in_world_fn, X_W_CB=X_W_RCB, tcp_transform=TCP_TRANSFORM
        )

        # Save  these so the Controllers can create their own planners
        self.inverse_kinematics_left_fn = inverse_kinematics_left_fn
        self.inverse_kinematics_right_fn = inverse_kinematics_right_fn
        self.is_state_valid_fn = collision_checker.CheckConfigCollisionFree

        planner = DualArmOmplPlanner(
            is_state_valid_fn=collision_checker.CheckConfigCollisionFree,
            inverse_kinematics_left_fn=inverse_kinematics_left_fn,
            inverse_kinematics_right_fn=inverse_kinematics_right_fn,
            joint_bounds_left=self.joint_bounds_left,
            joint_bounds_right=self.joint_bounds_right,
        )
        return planner


if __name__ == "__main__":
    import cv2
    import rerun as rr
    from airo_camera_toolkit.utils.image_converter import ImageConverter
    from cloth_tools.visualization.opencv import draw_pose

    # Check whether all hardware is connected
    station = CompetitionStation()
    camera = station.camera
    dual_arm = station.dual_arm

    X_W_C = station.camera_pose
    X_W_LCB = station.left_arm_pose
    X_W_RCB = station.right_arm_pose

    window_name = "Competiton station"
    cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(window_name, 1600, 800)

    rr.init("Competition Station - Point cloud", spawn=True)

    while True:
        image_rgb = camera.get_rgb_image_as_int()
        image_right_rgb = camera._retrieve_rgb_image_as_int(StereoRGBDCamera.RIGHT_RGB)
        point_cloud = camera._retrieve_colored_point_cloud()
        confidence_map = camera._retrieve_confidence_map()
        depth_map = camera._retrieve_depth_map()
        depth_image = camera._retrieve_depth_image()

        image_bgr = ImageConverter.from_numpy_int_format(image_rgb).image_in_opencv_format

        X_CB_LTCP = dual_arm.left_manipulator.get_tcp_pose()
        X_CB_RTCP = dual_arm.right_manipulator.get_tcp_pose()

        intrinsics = camera.intrinsics_matrix()
        draw_pose(image_bgr, np.identity(4), intrinsics, X_W_C, 0.25)
        draw_pose(image_bgr, X_W_LCB, intrinsics, X_W_C)
        draw_pose(image_bgr, X_W_RCB, intrinsics, X_W_C)
        draw_pose(image_bgr, X_W_LCB @ X_CB_LTCP, intrinsics, X_W_C, 0.05)
        draw_pose(image_bgr, X_W_RCB @ X_CB_RTCP, intrinsics, X_W_C, 0.05)

        cv2.imshow(window_name, image_bgr)
        key = cv2.waitKey(1)
        if key == ord("q"):
            break

        # Unfiltered point cloud in camera frame
        rr_point_cloud = rr.Points3D(positions=point_cloud.points, colors=point_cloud.colors)
        rr.log("world/point_cloud", rr_point_cloud)
        rr.log("image", rr.Image(image_rgb).compress(jpeg_quality=90))
        rr.log("image_right", rr.Image(image_right_rgb).compress(jpeg_quality=90))
        rr.log("depth_image", rr.Image(depth_image).compress(jpeg_quality=90))
        rr.log("depth_map", rr.DepthImage(depth_map))
        rr.log("confidence_map", rr.Image(confidence_map))



import numpy as np
from airo_typing import HomogeneousMatrixType, JointConfigurationType
from ur_analytic_ik import ur5e


def inverse_kinematics_in_world_fn(
    tcp_pose: HomogeneousMatrixType, X_W_CB: HomogeneousMatrixType, tcp_transform: HomogeneousMatrixType
) -> list[JointConfigurationType]:
    X_W_TCP = tcp_pose
    X_CB_W = np.linalg.inv(X_W_CB)
    solutions_1x6 = ur5e.inverse_kinematics_with_tcp(X_CB_W @ X_W_TCP, tcp_transform)
    solutions = [solution.squeeze() for solution in solutions_1x6]
    return solutions


# TODO document the functions below:
# in short: they change the solutions given by the IK that are within (-pi, pi)
# to lie between (-2pi, 2pi) of the angle of a joint is closer to the angle of a reference (e.g. starting) configuration
# This solved the issue of the robot arm "twisting" the cloth too much
# e.g. when for from -3.13 -> 1.57 instead of -3.13 -> -4.70
def post_process_solution(
    solution: JointConfigurationType, reference_configuration: JointConfigurationType
) -> JointConfigurationType:
    new_solution = np.zeros_like(solution)
    for i in range(len(solution)):
        angle = solution[i]
        alternative_angle = angle + 2 * np.pi if angle < 0 else angle - 2 * np.pi
        reference_angle = reference_configuration[i]
        abs_diff = np.abs(angle - reference_angle)
        abs_diff_new = np.abs(alternative_angle - reference_angle)
        new_solution[i] = alternative_angle if abs_diff_new < abs_diff else angle
    return new_solution


def inverse_kinematics_in_world_post_processed_fn(
    tcp_pose: HomogeneousMatrixType,
    X_W_CB: HomogeneousMatrixType,
    tcp_transform: HomogeneousMatrixType,
    reference_configuration: JointConfigurationType,
) -> list[JointConfigurationType]:
    X_W_TCP = tcp_pose
    X_CB_W = np.linalg.inv(X_W_CB)
    solutions_1x6 = ur5e.inverse_kinematics_with_tcp(X_CB_W @ X_W_TCP, tcp_transform)
    solutions = [solution.squeeze() for solution in solutions_1x6]
    solutions_post_processed = [post_process_solution(s, reference_configuration) for s in solutions]
    return solutions_post_processed



import numpy as np
from airo_camera_toolkit.point_clouds.operations import filter_point_cloud, generate_point_cloud_crop_mask
from airo_typing import BoundingBox3DType, NumpyDepthMapType, PointCloud, PointCloudPositionsType


def highest_point(points: PointCloudPositionsType) -> np.ndarray:
    return points[np.argmax(points[:, 2])]


def lowest_point(points: PointCloudPositionsType) -> np.ndarray:
    return points[np.argmin(points[:, 2])]


def filter_and_crop_point_cloud(
    point_cloud: PointCloud,
    confidence_map: NumpyDepthMapType,
    bounding_box: BoundingBox3DType,
    confidence_threshold: float = 1.0,
) -> PointCloud:
    """Filters and crops the point cloud to the region of interest.
    This function should be faster than doing both operations sequentially.

    Args:
        point_cloud: The point cloud to filter and crop.
        confidence_map: The confidence map (where high values are more uncertain).
        bounding_box: The bounding box.
        confidence_threshold: The confidence threshold.

    Returns:
        The filtered and cropped point cloud.
    """
    # Filter the low confidence points
    confidence_mask = (confidence_map <= confidence_threshold).reshape(-1)  # Threshold and flatten
    crop_mask = generate_point_cloud_crop_mask(point_cloud, bounding_box)

    # Combine the masks
    mask = confidence_mask & crop_mask

    point_cloud_filtered_and_cropped = filter_point_cloud(point_cloud, mask.nonzero())
    return point_cloud_filtered_and_cropped



import numpy as np
from airo_typing import HomogeneousMatrixType


def create_egocentric_world_frame(
    camera_pose_in_left: HomogeneousMatrixType, camera_pose_in_right: HomogeneousMatrixType
) -> tuple[HomogeneousMatrixType:, HomogeneousMatrixType, HomogeneousMatrixType]:
    """Creates an egocentric world frame based on the camera poses in the left and right robot base frames.

    This world frame is defined as follows:
    * The world origin is the middle point between the two robot bases.
    * The world Y-axis is the vector from the right robot base to the left robot base.
    * The world Z-axis is the same as the left robot base Z-axis (adjusted slightly to ensure orthogonality).
    * The world X-axis is the cross product of the world Y and Z axes.

    """
    X_LCB_C = camera_pose_in_left
    X_RCB_C = camera_pose_in_right
    X_LCB_RCB = X_LCB_C @ np.linalg.inv(X_RCB_C)
    p_LCB_RCB = X_LCB_RCB[:3, 3]

    # Define the world origin as the middle point between the two robot bases
    p_LCB_W = p_LCB_RCB / 2.0

    # Define the world Y-axis as the vector from the right robot base to the left robot base
    y_LCB_W = -p_LCB_RCB / np.linalg.norm(p_LCB_RCB)

    # Define (for now) that the world Z-axis is the same as the left_base Z-axis
    # We will fix later that it is not exactly orthogonal to the world Y-axis
    z_LCB_W = np.array([0, 0, 1])

    # Define the world X-axis as the cross product of the world Y and Z axes
    x_LCB_W = np.cross(y_LCB_W, z_LCB_W)
    x_LCB_W = x_LCB_W / np.linalg.norm(x_LCB_W)  # normalize because Z was not exactly orthogonal to Y

    # Ensure the world Z-axis is orthogonal to the world X and Y axes
    z_LCB_W = np.cross(x_LCB_W, y_LCB_W)

    # Create the rotation matrix
    R_LCB_W = np.column_stack((x_LCB_W, y_LCB_W, z_LCB_W))

    # Create the homogeneous transformation matrix
    X_LCB_W = np.identity(4)
    X_LCB_W[:3, :3] = R_LCB_W
    X_LCB_W[:3, 3] = p_LCB_W

    X_W_LCB = np.linalg.inv(X_LCB_W)
    X_W_RCB = X_W_LCB @ X_LCB_RCB

    X_W_C = X_W_LCB @ X_LCB_C
    # X_W_C_ = X_W_RCB @ X_RCB_C # Alternative, but should give exactly the same result I belive
    return X_W_C, X_W_LCB, X_W_RCB



import os

import requests


def upload_grasp(grasp_json_path: str, team_name: str, sample_id: str, server_url: str) -> None:
    """Uploads a grasp.json file to the Cloth Competition server.

    Args:
        grasp_json_path: Path to the grasp.json file.
        team_name: Your team's name, used to organize submissions.
        sample_id: The ID of the sample that the grasp was generated for, for example: 2024-04-22_09-26-45-537535
        server_url: URL of the server to upload to.

    """
    upload_url = server_url + "/upload_grasp"

    with open(grasp_json_path, "rb") as f:
        files = {"file": (os.path.basename(grasp_json_path), f)}
        data = {"team_name": team_name, "sample_id": sample_id}

        response = requests.post(upload_url, files=files, data=data)

    if response.status_code == 201:
        print(f"Grasp uploaded successfully: {response.json()}")
    else:
        print(f"Upload failed: {response.status_code}, {response.text}")



import os
import urllib
import zipfile
from pathlib import Path

import requests
from cloth_tools.dataset.format import COMPETITION_OBSERVATION_FILENAMES
from tqdm import tqdm


def download_and_extract_dataset(data_dir: str, dataset_zip_url: str) -> str:
    """Download and extract a Cloth Competition Dataset zip file from a URL.

    Args:
        data_dir: The parent directory to save the extracted dataset to.
        dataset_zip_url: The URL of the dataset zip file.

    Returns:
        The path to the extracted dataset directory.
    """
    os.makedirs(data_dir, exist_ok=True)
    with requests.get(dataset_zip_url, stream=True) as response:
        response.raise_for_status()

        total_size = int(response.headers.get("content-length", 0))
        block_size = 1024 * 1024  # 1 MB at a time

        zip_name = os.path.basename(urllib.parse.urlparse(dataset_zip_url).path)
        zip_path = os.path.join(data_dir, zip_name)
        dataset_dir = os.path.join(data_dir, zip_name.split(".")[0])

        progress_bar = tqdm(total=total_size, unit="B", unit_scale=True)

        with open(zip_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=block_size):
                if chunk:  # filter out keep-alive chunks
                    f.write(chunk)
                    progress_bar.update(len(chunk))

    with zipfile.ZipFile(zip_path) as z:
        z.extractall(data_dir)

    return dataset_dir


def get_latest_observation_start_url(server_url: str) -> str:
    url = server_url + "/latest_observation_start_dir"  # simple API to get the latest observation directory
    response = requests.get(url)
    return server_url + "/" + response.text


def download_latest_observation(dataset_dir: str, server_url: str) -> tuple[str, str]:
    """Download the latest start observation's files from the Cloth Competition server.
    Also returns the sample ID, which must be used when uploading a grasp.json file.

    Args:
        dataset_dir: The directory to save observation files to.
        server_url: URL of the server to download from.

    Returns:
        A tuple containing the path to the observation directory and the sample ID.
    """
    files_to_retrieve = COMPETITION_OBSERVATION_FILENAMES
    observation_dir_url = get_latest_observation_start_url(server_url)

    sample_dirname = observation_dir_url.split("/")[-2]
    sample_datetime = sample_dirname.split("sample_")[-1]

    sample_dir = Path(dataset_dir) / sample_dirname
    observation_start_dir = sample_dir / "observation_start"
    os.makedirs(observation_start_dir, exist_ok=True)

    for _, filename in files_to_retrieve.items():
        url = observation_dir_url + "/" + filename
        response = requests.get(url)
        filepath = os.path.join(observation_start_dir, filename)
        with open(filepath, "wb") as f:
            f.write(response.content)

    return str(observation_start_dir), sample_datetime



from typing import Optional, Tuple

import numpy as np
import open3d as o3d
from airo_typing import CameraIntrinsicsMatrixType, HomogeneousMatrixType


def open3d_camera(
    camera_pose: HomogeneousMatrixType,
    intrinsics: CameraIntrinsicsMatrixType,
    resolution: Tuple[int, int],
    color: Optional[Tuple[float, float, float]] is None,
    scale: float = 0.1,
) -> o3d.geometry.LineSet:
    """Creates a camera visualization in open3d.

    Args:
        camera_pose: camera pose
        intrinsics: camera intrinsics
        resolution: width, height of the camera's images

    Returns:
        camera_lines: a line set
    """

    camera_lines = o3d.geometry.LineSet.create_camera_visualization(
        *resolution, intrinsics, np.linalg.inv(camera_pose), scale=scale
    )
    if color is not None:
        camera_lines.paint_uniform_color(color)

    return camera_lines



import time

import cv2
import numpy as np
import scipy.ndimage
from airo_camera_toolkit.image_transforms.image_transform import ImageTransform
from airo_camera_toolkit.interfaces import RGBCamera
from airo_typing import OpenCVIntImageType
from loguru import logger


def calculate_variance_of_laplacian(image: OpenCVIntImageType):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    laplacian = cv2.Laplacian(gray, cv2.CV_64F)
    variance = laplacian.var()
    return variance


class MotionBlurDetector:
    def __init__(
        self,
        camera: RGBCamera,
        image_transform: ImageTransform,
        visualize=True,
        log_to_rerun: bool = True,
        threshold: float = 1.0,  # Ranges from > 200 to almost 0, 1 is almost motionless
    ):
        self.camera = camera
        self.image_transform = image_transform
        self.visualize = visualize
        self.log_to_rerun = log_to_rerun
        self.window_name = "Motion blur detection"

        self.variances = []
        self.variances_smoothed = []
        self.variances_of_variances_smoothed = []
        self.threshold = threshold

    def wait_for_blur_to_stabilize(self, warmup: float = 2.0, timeout: float = 10.0) -> bool:
        """Wait for the motion blur to stabilize.

        Args:
            timeout: the maximum time to wait for the motion blur to stabilize

        Returns:
            True if the motion blur stabilized within the timeout, False otherwise
        """
        time_start = time.time()
        time_last_log = 0.0

        if self.visualize:
            cv2.namedWindow(self.window_name, cv2.WINDOW_NORMAL)

        fps = self.camera.fps

        history = int(2 * fps)
        sigma = history + 1  # make it odd

        while True:
            image_rgb = self.camera.get_rgb_image_as_int()
            image_bgr = cv2.cvtColor(image_rgb, cv2.COLOR_RGB2BGR)
            image_transformed = self.image_transform.transform_image(image_bgr)

            variance = calculate_variance_of_laplacian(image_transformed)
            self.variances.append(variance)
            variance_smoothed = scipy.ndimage.gaussian_filter1d(self.variances, sigma=sigma)[-1]
            self.variances_smoothed.append(variance_smoothed)
            variance_of_variances_smoothed = np.var(self.variances_smoothed[-history:])
            self.variances_of_variances_smoothed.append(variance_of_variances_smoothed)

            if time.time() - time_last_log > 1.0:
                logger.info(
                    f"Variance of the smoothed variance of the Laplacian: {variance_of_variances_smoothed:.2f}"
                )
                time_last_log = time.time()

            if self.log_to_rerun:
                import rerun as rr

                rr.log("laplacian_variance/plot", rr.Scalar(variance_smoothed))
                rr.log("variance_of_variance/plot", rr.Scalar(variance_of_variances_smoothed))

            if variance_of_variances_smoothed < self.threshold and time.time() - time_start > warmup:
                logger.info("Motion blur stabilized")
                cv2.destroyWindow(self.window_name)
                return True

            if time.time() - time_start > timeout:
                logger.warning(
                    f"Motion blur did not stabilize within the timeout. Var: {variance_of_variances_smoothed:.2f}"
                )
                cv2.destroyWindow(self.window_name)
                return False

            if self.visualize:
                cv2.imshow(self.window_name, image_transformed)
                key = cv2.waitKey(1)
                if key == ord("q"):
                    cv2.destroyWindow(self.window_name)
                    return False


# if __name__ == "__main__":
#     camera_kwargs = {
#         "resolution": Zed2i.RESOLUTION_2K,
#         "depth_mode": Zed2i.NEURAL_DEPTH_MODE,
#         "fps": 15,
#     }

#     camera =Zed2i(**camera_kwargs)
#     image_transform = ImageTransform()
#     motion_blur_detector = MotionBlurDetector(camera, image_transform)
#     motion_blur_detector.wait_for_blur_to_stabilize()



import tempfile

import airo_models
import numpy as np
import open3d as o3d
from airo_camera_toolkit.point_clouds.conversions import point_cloud_to_open3d
from airo_drake import X_URBASE_ROSBASE, add_floor, add_manipulator, add_wall
from airo_models import mesh_urdf_path
from airo_typing import HomogeneousMatrixType, PointCloud
from pydrake.math import RigidTransform, RollPitchYaw
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.planning import RobotDiagramBuilder

# Some fixed transforms and default poses (e.g. for when you don't have access to the real robot)
ARM_Y_DEFAULT = 0.45

X_CB_B = X_URBASE_ROSBASE

# The default pose of the left robot base when using the ROS URDFs
X_W_L_DEFAULT = RigidTransform(rpy=RollPitchYaw([0, 0, np.pi / 2]), p=[0, ARM_Y_DEFAULT, 0])
X_W_LCB_DEFAULT = X_W_L_DEFAULT.GetAsMatrix4() @ np.linalg.inv(X_CB_B.GetAsMatrix4())

# The default pose of the right robot base when using the ROS URDFs
X_W_R_DEFAULT = RigidTransform(rpy=RollPitchYaw([0, 0, np.pi / 2]), p=[0, -ARM_Y_DEFAULT, 0])
X_W_RCB_DEFAULT = X_W_R_DEFAULT.GetAsMatrix4() @ np.linalg.inv(X_CB_B.GetAsMatrix4())


def add_cloth_competition_dual_ur5e_scene(
    robot_diagram_builder: RobotDiagramBuilder,
    X_W_LCB: HomogeneousMatrixType | None = None,
    X_W_RCB: HomogeneousMatrixType | None = None,
) -> tuple[tuple[ModelInstanceIndex, ModelInstanceIndex], tuple[ModelInstanceIndex, ModelInstanceIndex]]:
    parser = robot_diagram_builder.parser()
    parser.SetAutoRenaming(True)

    X_W_L = X_W_L_DEFAULT if X_W_LCB is None else RigidTransform(X_W_LCB @ X_CB_B.GetAsMatrix4())
    X_W_R = X_W_R_DEFAULT if X_W_RCB is None else RigidTransform(X_W_RCB @ X_CB_B.GetAsMatrix4())

    add_floor(robot_diagram_builder, y_size=2.4)

    # Add three safety walls
    arm_y = X_W_L.translation()[1]
    wall_thickness = 0.2
    wall_left_position = np.array([0, arm_y + 0.7 + wall_thickness / 2, 0])
    wall_right_position = np.array([0, -arm_y - 0.7 - wall_thickness / 2, 0])
    wall_back_position = np.array([0.9 + wall_thickness / 2, 0, 0])
    add_wall(robot_diagram_builder, "XZ", 2.0, 2.0, wall_thickness, wall_left_position, "wall_left")
    add_wall(robot_diagram_builder, "XZ", 2.0, 2.0, wall_thickness, wall_right_position, "wall_right")
    add_wall(robot_diagram_builder, "YZ", 2.0, 2.7, wall_thickness, wall_back_position, "wall_back")

    # The robot arms
    arm_left_index, gripper_left_index = add_manipulator(
        robot_diagram_builder, "ur5e", "robotiq_2f_85", X_W_L, static_gripper=True
    )
    arm_right_index, gripper_right_index = add_manipulator(
        robot_diagram_builder, "ur5e", "robotiq_2f_85", X_W_R, static_gripper=True
    )

    return (arm_left_index, arm_right_index), (gripper_left_index, gripper_right_index)


def add_cloth_obstacle_to_builder(
    robot_diagram_builder: RobotDiagramBuilder, point_cloud_cloth: PointCloud
) -> tuple[ModelInstanceIndex, o3d.t.geometry.TriangleMesh]:

    # Create mesh URDF of the convex hull of the cloth point cloud
    pcd = point_cloud_to_open3d(point_cloud_cloth)
    hull = pcd.compute_convex_hull()

    # Scaling disable for now because this might make it collide with that gripper that's holding it in the air
    # hull.scale(1.1, hull.get_center())  # make 10% larger

    hull_file = tempfile.NamedTemporaryFile(prefix="cloth_hull_", suffix=".obj", delete=False)
    hull_filename = hull_file.name
    o3d.t.io.write_triangle_mesh(hull_filename, hull)  # equiv? hull.write_obj(hull_filename)

    hull_urdf_path = mesh_urdf_path(hull_filename, "cloth_hull")

    # Add the cloth hull to the scene
    plant = robot_diagram_builder.plant()
    parser = robot_diagram_builder.parser()
    hull_index = parser.AddModels(hull_urdf_path)[0]

    world_frame = plant.world_frame()
    hull_frame = plant.GetFrameByName("base_link", hull_index)
    plant.WeldFrames(world_frame, hull_frame)

    return hull_index, hull


def add_safety_wall_to_builder(
    robot_diagram_builder: RobotDiagramBuilder, X_W_TCP: HomogeneousMatrixType
) -> ModelInstanceIndex:
    plant = robot_diagram_builder.plant()
    parser = robot_diagram_builder.parser()
    safety_wall_thickness = 0.05
    safety_wall_urdf_path = airo_models.box_urdf_path((0.2, safety_wall_thickness, 0.3), "safety_wall")

    # Position the safety wall in front of the left TCP
    shift = safety_wall_thickness / 2 + 0.01
    p_W_S = X_W_TCP[:3, 3] + X_W_TCP[:3, 2] * shift
    X_W_S = RigidTransform(p=p_W_S)

    # Add the cloth hull to the scene
    safety_wall_index = parser.AddModels(safety_wall_urdf_path)[0]
    world_frame = plant.world_frame()
    safety_wall_frame = plant.GetFrameByName("base_link", safety_wall_index)
    plant.WeldFrames(world_frame, safety_wall_frame, X_W_S)

    return safety_wall_index



import json
import os
from dataclasses import dataclass
from typing import Tuple

import cv2
import numpy as np
from airo_camera_toolkit.pinhole_operations.projection import project_points_to_image_plane
from airo_camera_toolkit.pinhole_operations.unprojection import unproject_onto_depth_values
from airo_camera_toolkit.point_clouds.conversions import point_cloud_to_open3d
from airo_camera_toolkit.utils.image_converter import ImageConverter
from airo_dataset_tools.data_parsers.pose import Pose
from airo_spatial_algebra import transform_points
from airo_typing import (
    CameraIntrinsicsMatrixType,
    HomogeneousMatrixType,
    NumpyDepthMapType,
    NumpyIntImageType,
    OpenCVIntImageType,
    PointCloud,
    Vector3DType,
)
from cloth_tools.visualization.opencv import draw_pose
from loguru import logger
from pydantic import BaseModel


def top_down_camera_pose(height: float = 1.5) -> np.ndarray:
    Z = np.array([0, 0, -1])
    X = np.array([0, -1, 0])  # X is the rightwards direction for the camera
    Y = np.cross(Z, X)
    orientation = np.column_stack([X, Y, Z])
    position = np.array([0, 0, height])

    pose = np.identity(4)
    pose[:3, :3] = orientation
    pose[:3, 3] = position

    return pose


def grasp_hanging_cloth_pose(
    position: Vector3DType, approach_direction: Vector3DType, grasp_depth: float = 0.0
) -> HomogeneousMatrixType:
    """Create a pose for grasping a hanging cloth.

    Args:
        position: The position of the grasp.
        approach_direction: The direction the gripper should approach the grasp from.
        grasp_depth: An additional forward (= Z) offset for the grasp to bring it deeper into the cloth.

    Returns:
        pose: The pose for the grasp.
    """
    Z = approach_direction / np.linalg.norm(approach_direction)
    position_with_depth = position + grasp_depth * Z

    # Pointing gripper Y up or down leads to the gripper opening horizontally
    # I chose Y-down here because it's closer the other poses used in the controllers
    # Y = np.array([0, 0, -1])  # default Y
    Y = np.array([0, 0, 1])  # Trying the positive instead to attempt to reduce twisting issue

    # Handle rare case where Z is parallel to default Y
    if np.abs(np.dot(Y, Z)) > 0.99:
        Y = np.array([-1, 0, 0])

    X = np.cross(Y, Z)
    X = X / np.linalg.norm(X)  # Normalize X for the case where Y and Z were not perpendicular

    # Recalculate Y to be guaranteed perpendicular to X and Z
    Y = np.cross(Z, X)

    orientation = np.column_stack([X, Y, Z])

    pose = np.identity(4)
    pose[:3, :3] = orientation
    pose[:3, 3] = position_with_depth

    return pose


def project_point_cloud_to_image(
    point_cloud: PointCloud,
    camera_pose: HomogeneousMatrixType,
    intrinsics: np.ndarray,
    resolution: Tuple[int, int],
    background_color: Tuple[int, int, int] = (0, 0, 0),
    blur_image: bool = False,
) -> OpenCVIntImageType:
    pcd = point_cloud_to_open3d(point_cloud)
    rgbd_image = pcd.project_to_rgbd_image(*resolution, intrinsics, extrinsics=np.linalg.inv(camera_pose))

    image_rgb_float = np.asarray(rgbd_image.color)
    depth_map_float = np.asarray(rgbd_image.depth).squeeze()

    # Convert to uint8
    image = ImageConverter.from_numpy_format(image_rgb_float).image_in_opencv_format

    # Turn pixels with zero depth to gray
    image[depth_map_float == 0.0] = background_color

    if blur_image:
        image = cv2.blur(image, (3, 3))
        image = cv2.medianBlur(image, 3)

    return image


def calculate_approach_direction_from_clicked_points(
    frontal_clicked_in_world: Vector3DType,
    topdown_clicked_in_image: Tuple[int, int],
    camera_topdown_pose: HomogeneousMatrixType,
    intrinsics: CameraIntrinsicsMatrixType,
):
    p_W = frontal_clicked_in_world
    X_W_VC = camera_topdown_pose
    topdown_clicked_in_image = np.array(topdown_clicked_in_image)

    X_VC_W = np.linalg.inv(X_W_VC)
    frontal_clicked_in_image = project_points_to_image_plane(transform_points(X_VC_W, p_W), intrinsics).squeeze()

    # Image space vector from topdown to frontal clicked points
    vector_in_image = frontal_clicked_in_image - topdown_clicked_in_image

    # Convert to camera space 3D vector
    vector_in_camera = np.array([vector_in_image[0], vector_in_image[1], 0])  # Image space to camera space vector
    vector_in_camera = vector_in_camera / np.linalg.norm(vector_in_camera)  # Normalize to unit vector

    # Transform to world space vector
    v_VC = vector_in_camera
    R_W_VC = X_W_VC[:3, :3]
    v_W = R_W_VC @ v_VC

    approach_direction = v_W
    return approach_direction


def calculate_grasp_pose_from_annotations(
    frontal_clicked_in_image: Tuple[int, int],
    topdown_clicked_in_image: Tuple[int, int] | None,
    grasp_depth: float,
    depth_map: NumpyDepthMapType,
    camera_pose: HomogeneousMatrixType,
    camera_topdown_pose: HomogeneousMatrixType,
    intrinsics: CameraIntrinsicsMatrixType,
) -> HomogeneousMatrixType:
    X_W_C = camera_pose
    X_W_VC = camera_topdown_pose

    # Transforming the point clicked in the frontal view to the world frame
    u, v = frontal_clicked_in_image
    image_coordinates = np.array([u, v]).reshape((1, 2))
    depth_values = np.array([depth_map[v, u]])
    p_C = unproject_onto_depth_values(image_coordinates, depth_values, intrinsics)
    p_W = transform_points(X_W_C, p_C).squeeze()

    # Calculate the approach direction
    Z = np.array([1, 0, 0])

    if topdown_clicked_in_image is not None:
        Z = calculate_approach_direction_from_clicked_points(p_W, topdown_clicked_in_image, X_W_VC, intrinsics)

    grasp_pose = grasp_hanging_cloth_pose(p_W, Z, grasp_depth)
    return grasp_pose


def draw_line_to_point_in_world(
    image: OpenCVIntImageType,
    point_in_image: Tuple[int, int],
    point_in_world: Vector3DType,
    camera_pose: HomogeneousMatrixType,
    intrinsics: CameraIntrinsicsMatrixType,
    color: Tuple[int, int, int] = (0, 255, 0),
) -> OpenCVIntImageType:
    point_in_world_projected = project_points_to_image_plane(
        transform_points(camera_pose, point_in_world), intrinsics
    ).squeeze()
    point_in_world_projected_int = tuple(np.rint(point_in_world_projected).astype(int))
    cv2.line(image, point_in_image, point_in_world_projected_int, color, 1, cv2.LINE_AA)


@dataclass
class GraspAnnotationInfo:
    """Some additional information about the grasp annotation e.g. for visualization."""

    grasp_pose: HomogeneousMatrixType | None
    clicked_point_frontal: Tuple[int, int] | None
    clicked_point_topdown: Tuple[int, int] | None
    grasp_depth: float
    image_frontal: NumpyIntImageType
    image_topdown: NumpyIntImageType


def get_manual_grasp_annotation(  # noqa: C901
    image: NumpyIntImageType,
    depth_map: NumpyDepthMapType,
    point_cloud: PointCloud,
    camera_pose: HomogeneousMatrixType,
    intrinsics: CameraIntrinsicsMatrixType,
    log_to_rerun: bool = False,
) -> GraspAnnotationInfo | None:
    """Manually annotate a grasp pose on a hanging piece of cloth.

    Args:
        image: A (frontal) image.
        depth_map: The corresponding depth map.
        point_cloud: The (filtered) point cloud in the world frame.
        camera_pose: The pose of the camera in the world frame.
        intrinsics: The camera intrinsics.
    """
    cyan = (255, 255, 0)
    yellow = (0, 255, 255)

    window_frontal = "Grasp Annotation - Frontal view"
    window_topdown = "Grasp Annotation - Topdown view"

    logger.info("[Grasp Annotation] - Usage:")
    logger.info("Click in the front view to set grasp location.")
    logger.info("Click in the topdown view to set the approach direction.")
    logger.info("Press 'arrow up' to increase the grasp depth.")
    logger.info("Press 'arrow down' to decrease the grasp depth.")
    logger.info("Press 'y' to confirm the grasp pose.")
    logger.info("Press 'b' to toggle blur in the topdown view.")
    logger.info("Press 'q' to quit.")

    h = 600
    w = int(1.5 * h)
    cv2.namedWindow(window_frontal, cv2.WINDOW_NORMAL)
    cv2.namedWindow(window_topdown, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(window_frontal, w, h)
    cv2.resizeWindow(window_topdown, w, h)
    cv2.moveWindow(window_frontal, 0, 0)
    cv2.moveWindow(window_topdown, w + 80, 0)

    clicked_point = {window_frontal: None, window_topdown: None}

    def callback_frontal(event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            logger.info(f"Frontal view: click at ({x}, {y}), depth = {depth_map[y, x]:.3f} meter")
            clicked_point[window_frontal] = (x, y)

        if event == cv2.EVENT_RBUTTONDOWN:
            logger.info("Frontal view: removed click.")
            clicked_point[window_frontal] = None

    def callback_topdown(event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            logger.info(f"Topdown view: click at ({x}, {y})")
            clicked_point[window_topdown] = (x, y)
        if event == cv2.EVENT_RBUTTONDOWN:
            logger.info("Topdown view: removed click.")
            clicked_point[window_topdown] = None

    cv2.setMouseCallback(window_frontal, callback_frontal)
    cv2.setMouseCallback(window_topdown, callback_topdown)

    if log_to_rerun:
        import rerun as rr

        rr.init("Grasp Annotation - Point cloud", spawn=True)
        rr.log("world/point_cloud", rr.Points3D(positions=point_cloud.points, colors=point_cloud.colors))
        rr.log("world/identity", rr.Transform3D(scale=0.5))

    X_W_C = camera_pose
    X_W_VC = top_down_camera_pose(height=1.5)

    resolution = (image.shape[1], image.shape[0])

    image_frontal = ImageConverter.from_numpy_int_format(image).image_in_opencv_format
    image_topdown_sharp = project_point_cloud_to_image(
        point_cloud, X_W_VC, intrinsics, resolution, background_color=(90, 90, 90)
    )
    image_topdown_blurred = project_point_cloud_to_image(
        point_cloud, X_W_VC, intrinsics, resolution, background_color=(90, 90, 90), blur_image=True
    )

    image_topdown = image_topdown_blurred
    blur_image = True
    grasp_pose = None
    grasp_depth = 0.05

    while True:
        image_frontal_annotated = image_frontal.copy()
        image_topdown_annotated = image_topdown.copy()

        if clicked_point[window_topdown] is not None:
            cv2.circle(image_topdown_annotated, clicked_point[window_topdown], 5, cyan, 3, cv2.LINE_AA)

        if clicked_point[window_frontal] is not None:
            cv2.circle(image_frontal_annotated, clicked_point[window_frontal], 5, yellow, 3, cv2.LINE_AA)

            grasp_pose = calculate_grasp_pose_from_annotations(
                clicked_point[window_frontal],
                clicked_point[window_topdown],
                grasp_depth,
                depth_map,
                X_W_C,
                X_W_VC,
                intrinsics,
            )

            draw_pose(image_frontal_annotated, grasp_pose, intrinsics, X_W_C)
            draw_pose(image_topdown_annotated, grasp_pose, intrinsics, X_W_VC)

            if clicked_point[window_topdown] is not None:
                draw_line_to_point_in_world(
                    image_topdown_annotated, clicked_point[window_topdown], grasp_pose[:3, 3], X_W_VC, intrinsics, cyan
                )

            if log_to_rerun:
                rr.log("world/grasp_pose", rr.Transform3D(translation=grasp_pose[:3, 3], mat3x3=grasp_pose[:3, :3]))

        text = f"Grasp depth: {grasp_depth:.2f} m (Press d/f to change)"
        cv2.putText(image_frontal_annotated, text, (10, 60), cv2.FONT_HERSHEY_SIMPLEX, 2, (0, 255, 0), 3, cv2.LINE_AA)

        cv2.imshow(window_frontal, image_frontal_annotated)
        cv2.imshow(window_topdown, image_topdown_annotated)
        key = cv2.waitKey(10)

        grasp_annotation_info = GraspAnnotationInfo(
            grasp_pose=grasp_pose,
            clicked_point_frontal=clicked_point[window_frontal],
            clicked_point_topdown=clicked_point[window_topdown],
            grasp_depth=grasp_depth,
            image_frontal=image_frontal_annotated.copy(),
            image_topdown=image_topdown_annotated.copy(),
        )

        if key == ord("q") or key == ord("n"):
            # Signal that the grasp annotation was aborted
            grasp_pose = None
            cv2.destroyWindow(window_frontal)
            cv2.destroyWindow(window_topdown)
            return grasp_annotation_info
        if key == ord("y"):
            cv2.destroyWindow(window_frontal)
            cv2.destroyWindow(window_topdown)
            return grasp_annotation_info
        if key == ord("b"):
            blur_image = not blur_image
            image_topdown = image_topdown_blurred if blur_image else image_topdown_sharp
        if key == ord("d"):  # deeper into the cloth
            grasp_depth += 0.01
        if key == ord("f"):  # further away from the cloth
            grasp_depth -= 0.01


class GraspAnnotation(BaseModel):
    clicked_point_frontal: Tuple[int, int]
    clicked_point_topdown: Tuple[int, int] | None = None
    grasp_depth: float


def save_grasp_info(dir: str, grasp_info: GraspAnnotationInfo):

    os.makedirs(dir, exist_ok=True)

    grasp_pose_file = os.path.join(dir, "grasp_pose.json")

    with open(grasp_pose_file, "w") as f:
        grasp_pose_model = Pose.from_homogeneous_matrix(grasp_info.grasp_pose)
        json.dump(grasp_pose_model.model_dump(exclude_none=False), f, indent=4)

    grasp_annotation_file = os.path.join(dir, "grasp_annotation.json")
    grasp_annotation = GraspAnnotation(
        clicked_point_frontal=grasp_info.clicked_point_frontal,
        clicked_point_topdown=grasp_info.clicked_point_topdown,
        grasp_depth=grasp_info.grasp_depth,
    )

    with open(grasp_annotation_file, "w") as f:
        json.dump(grasp_annotation.model_dump(exclude_none=True), f, indent=4)

    # Save the two images with the grasp visualized
    grasp_frontal_image_file = os.path.join(dir, "frontal_image_grasp.jpg")
    grasp_topdown_image_file = os.path.join(dir, "topdown_image_grasp.jpg")
    cv2.imwrite(grasp_frontal_image_file, grasp_info.image_frontal)
    cv2.imwrite(grasp_topdown_image_file, grasp_info.image_topdown)
